unit TermInterpreter;

interface

var LastErrorMsg : String;
    LastErrorPos : Integer;
    Okay         : Boolean;
function TermWert(s: String): Double;

implementation

uses math;

{ Interne Hilfsfunktionen }

function GetFirstChar(var s: String): Char;
  begin
  While (Length(s) > 0) and (s[1] in [' ']) do
    Delete(s, 1, 1);             { killt fhrende Leerzeichen        }
  If Length(s) > 0 then
    Result := s[1]               { gibt das erste Zeichen zurck.... }
  else
    Result := #0;                { ... oder bei leerem String ein #0 }
  end;

{ Scannerfunktionen fr logische Term-Bestandteile }

function N_Zahl(var s: String; var res: Double): Boolean;
  var c : Char;
  begin
  c := GetFirstChar(s);
  If c in ['0'..'9'] then begin  { Es wird eine Ziffer erwartet....    }
    Result := True;
    res := 0;
    Repeat                       { Solange Ziffern da sind, werden sie }
      Delete(s, 1, 1);
      res := 10 * res + (Ord(c) - Ord('0'));  { ... ausgewertet.       }
      c := GetFirstChar(s);                   { Nchstes Zeichen holen }
    until not (c in ['0'..'9']); { bis keine Ziffer mehr gefunden wird }
    end
  else begin                     { Falls keine Ziffer da ist: Fehler!  }
    Result := False;
    LastErrorMsg := 'Ziffer erwartet!';       { Fehlermeldung absetzen }
    end;
  end;

function Zahl(var s: String; var res: Double): Boolean;
  var IsMinus,
      IsMinusExp : Boolean;
      c          : Char;
      d, e       : Double;
  begin
  IsMinus := False;              { Normalerweise erwarten wir was Positives! }
  c := GetFirstChar(s);
  If c in ['+', '-'] then begin        { Optionales Vorzeichen testen }
    Delete(s, 1, 1);
    IsMinus := c = '-';                { Gegebenenfalls auf Minus umschalten }
    end;
  If N_Zahl(s, res) then begin         { Vorkomma-Anteil holen }
    Result := True;                    { => Erfolg ! }
    c := GetFirstChar(s);
    If c in ['.', ','] then begin      { Nachkommastellen vorhanden ? }
      Delete(s, 1, 1);                 { Dezimaltrenner lschen }
      c := GetFirstChar(s);
      If c in ['0'..'9'] then begin    { Ziffern einlesen }
        d := 0.1;                      { Stellenwert initialisieren }
        Repeat
          Delete(s, 1, 1);
          res := res + d * (Ord(c) - Ord('0'));  { Wert berechnen }
          d   := d / 10;               { Stellenwert fr nchste Stelle }
          c   := GetFirstChar(s);      { Nchste Ziffer holen ... }
        until not (c in ['0'..'9']);   { ...bis keine mehr da sind }
        end
      else begin
        Result := False;
        LastErrorMsg := 'Nachkommastellen erwartet!';
        end;
      end;
    If Result and                      { Bisher kein Fehler ! }
       (c in ['e', 'E']) then begin    { Exponent vorhanden ? }
      Delete(s, 1, 1);                 { 'e' lschen }
      IsMinusExp := False;
      c := GetFirstChar(s);
      If c in ['+', '-'] then begin    { optionales Vorzeichen testen }
        Delete(s, 1, 1);
        IsMinusExp := c = '-';         { gegebenenfalls auf Minus umschalten }
        end;
      If N_Zahl(s, e) then begin       { Exponent holen }
        If IsMinusExp then e := -e;
        res := res * power(10, e);
        end
      else begin                       { Kein Exponent vorhanden ? => Fehler! }
        Result := False;
        LastErrorMsg := 'Exponent erwartet!';
        end;
      end;
    If IsMinus then res := -res;       { Erst hier ber das VZ entscheiden !!! }
    end
  else
    Result := False;             { LastErrorMsg wird schon von N_Zahl gesetzt. }
  end;

function Summe(var s: String; var res: Double): Boolean;
  var r1 : Double;
      c  : Char;
  begin
  If Zahl(s, res) then begin       { Ersten Summanden holen                  }
    Result := True;
    c := GetFirstChar(s);
    While c in ['+', '-'] do begin { Solange eine Strichrechnung folgt :     }
      Delete(s, 1, 1);                 { Rechenzeichen lschen               }
      If Zahl(s, r1) then begin        { Nchsten Summanden holen...         }
        If c = '+' then
          res := res + r1              { und zur aktuellen Summe addieren... }
        else
          res := res - r1;             { ...bzw. von ihr subtrahieren        }
        c := GetFirstChar(s);
        end
      else begin                       { Kein Summand vorhanden ? => Fehler! }
        Result := False;               {  LastErrorMsg wird in Zahl gesetzt  }
        c      := #0;                  { While-Schleife verlassen }
        end;
      end;
    end
  else                             { Kein Summand da ? => Fehler!            }
    Result := False;               {     LastErrorMsg wird in Zahl gesetzt   }
  end;


{ Interface-Funktion }

function TermWert(s: String): Double;
  { Diese Funktion dient hauptschlich der Fehlerverwaltung. Sie sorgt dafr,
    dass bei einem Fehler whrend des Interpretationsvorgangs die Variablen
    LastErrorMsg und LastErrorPos korrekt gesetzt werden. Auerdem verwaltet
    sie die Variable Okay, in der ber den Erfolg des Interpretationsversuchs
    Buch gefhrt wird. }
  var r   : Double;
      ils : Integer;
  begin
  Okay  := True;
  ils   := Length(s);
  LastErrorMsg := '';
  LastErrorPos := ils;
  IF Summe(s, r) then begin  { Hier wird gerechnet: jeder Term ist eine Summe. }
    Result := r;
    Okay   := True;
    end
  else begin                    { Buchhaltung im Fehlerfall: }
    Result := 0;
    Okay := False;
    LastErrorPos := ils - Length(s);
    LastErrorMsg := LastErrorMsg + '  {Rest : ' + s + ' }';
    end;
  end;

end.
